/*
  /-------------------------------------------------\
  |                                                 |
  |                                                 |
  |     /=====\     /=====\     |\          /|      |
  |    /       \   /       \    | \        / |      |
  |    |           |            |  \      /  |      |
  |    |           |            |   \    /   |      |
  |    \       /   \       /    |    \  /    |      |
  |     \=====/     \=====/     |     \/     |      |
  |                                                 |
  |                                                 |
  \-------------------------------------------------/

  ===================================================

  Name:       CCM (define.C)
  
  Written by: Damian J. J. Farnell

  Date:       Wed Sep 29 1999
 
  Purpose:    Determine and solve CCM LSUBm equations.

  ===================================================

  ---------------------------------------------------

  MODULE NAME: DEFINE.C 

  PURPOSE:     Defines the LSUBm clusters recursively.
               Not optimised. 

  ---------------------------------------------------

*/

#include<iostream.h>
#include<stdlib.h>
#include<stdlib.h>
#include<math.h>
#include<stdio.h>
#include<string.h>
#if (PARALLEL_FLAG>0)
#include<mpi.h>
#endif

#include"../include/nr.h"
#include"../include/nrutil.h"
#include"../include/Constants.h"
#include"../include/Externals.h"
#include"../include/FunctionPrototypes.h"

void define(void)
{

  /* this module defines all of the cluster configurations using
     a recursive algorithm. 

     In essence this is a souped-up version of the algorithm used
     in the original ``define configurations'' C program. Here, for
     complex crystallogrphic lattices and higher spin.

  */

  int i,j,k,l,m,n,tempLatticeSymmetries;
  int cell;

  int NumberConnected=0;
  int NumberDisconnected=0;
  int restriction;

  char path[]="../Data/";
  char configurations[80], configs_name[]=".configs";

  int **configuration;
  int descriptor[MAX_SPIN_LIST];
  int length,descriptorIndex;

  int **tempConfig;
  tempConfig=matrix(0,MAX_LSUBM+1,0,5);

  
  strcpy(configurations,Control.name);
  strcat(configurations,configs_name);
  
  fp1=fopen(configurations,"w");  
  if (!fp1) {
    cout << "Cannot Open " << configurations << "...\nExiting!!\n\n";
    exit(1);
  }

  configuration=matrix(0,MAX_LSUBM+1,0,5);

  if (Control.output > 0 && myid == 0) {
    cout << "\n============================================================\n";
    printf( "             DEFINE CLUSTERS: START ALGORITHM.\n");
    cout << "============================================================\n\n";
  }

  /* !! NEW ALGORITHM FOR S>= 1/2.... 30-8-2000 */

  /* PART 1: DETERMINE THE LATTICE ANIMALS */

  printf("PART 1: DETERMINE THE LATTICE ANIMALS \n\n");

  NumberLatticeAnimals=0;
  NumberConfigurations=0;

  tempLatticeSymmetries=LatticeSymmetries.number;
  LatticeSymmetries.number=1;

  for(cell=1;cell<=UnitCell.number;cell++) {
    length=1;
    configuration[1][1]=0;
    configuration[1][2]=0;
    configuration[1][3]=0;
    configuration[1][4]=cell;      
    recursiveClusteringFunction(configuration,
				length,
				Params.m,
				NumberLatticeAnimals);
  }

  /* PRINT OUT INFORMATION ON CLUSTERS */

  if (Control.output > 2 && myid == 0) {
    for(i=1;i<=NumberLatticeAnimals;i++) {
      printOutClusterFromDescriptor(latticeAnimals[i].descriptor);
    }
  }
   
  printf("\nDEFINE LATTICE ANIMALS: COMPLETED SUCCESSFULLY.\n\n");

  /* PART 2: DETERMINE THE FUNDAMTENTAL SPIN CONFIGURATIONS */
  
  printf("\nPART 2: DEFINE FUNDAMENTAL SPIN CONFIGURATIONS.\n\n");

  LatticeSymmetries.number=tempLatticeSymmetries;
  findFundamentalSpinConfigurations();

  printf("\nDEFINE FUND. SPIN CONFIGS.: COMPLETED SUCCESSFULLY.\n\n");

  /* PART 3: PRINT OUT FUNDAMENTAL SPIN CONFIGURATIONS */
  
  printf("\nPART 3: PRINT OUT FUNDAMENTAL SPIN CONFIGURATIONS.\n\n"); 
  
  if (Control.output > 0 && myid == 0) {
    for(i=1;i<=NumberConfigurations;i++) {
      printf("Cluster Number (%d): \n",i);
      printOutDescriptor(spinConfigurations[i].descriptor);
      printOutClusterFromDescriptor(spinConfigurations[i].descriptor);
      printf("\n");
    }
  }
  cout <<endl;
  cout << "Number of Connected Clusters    = " << NumberConnected << endl;
  cout << "Number of Disconnected Clusters = " << NumberDisconnected << endl;
  cout << "Total Number of Clusters        = " << NumberConfigurations << endl;
  cout << endl <<endl;

  fwrite(&NumberConfigurations,sizeof(int),1,fp1);
  for(i=1;i<=NumberConfigurations;i++) {
    fwrite(spinConfigurations[i].descriptor,
	   ((MAX_SPIN_LIST)*sizeof(int)),
	   1,
	   fp1);
  }

  fclose(fp1);

   /* NEW!!: work out the maximum radius for all clusters 
     This is used to reduce the amount of sweeping for 
     F_k F_m s_k^z s_m^z -like terms. */

  radius=0;

  for(i=1;i<=NumberConfigurations;i++) {

    length=findClusterFromDescriptor(spinConfigurations[i].descriptor,
				     tempConfig);

    /* note that midPointX, midPointY, midPointZ indicate Unit cells */

    // note that integer rounding shouldn't affect this 
    // algorithm

    // now find the maximum `radius'

    for(j=1;j<=length;j++) {
      if (abs(tempConfig[j][1]) > radius)
	radius=abs(tempConfig[j][1]);
      if (abs(tempConfig[j][2]) > radius)
	radius=abs(tempConfig[j][2]);
      if (abs(tempConfig[j][3]) > radius)
	radius=abs(tempConfig[j][3]);
    }
  }

  radius+=1;

  printf("\nRadius = %d\n\n",radius);

  printf("DEFINE CLUSTERS: ALGORITHM SUCCESFULLY COMPLETED!\n\n");

}

void recursiveClusteringFunction(int **configuration, 
				 int length, 
				 int maxorder, 
				 int &NumberAnimals) {

  /* A simple recursive function that determines all connected configs */

  int tempLength;
  int i,j,k,l;
  int descriptorIndex;
  int numberFlips;
  int maxNumberFlips;
  int index;
  int tempx,tempy,tempz,tempCell;

  int descriptor[MAX_SPIN_LIST];
  
  // first check whether we have an error using the `length' of cluster. 

  if (length > maxorder) {
    cout << "Error in Recursion: length > maxorder!\nExiting...\n\n";
    exit(1); 
  } 

  // first check if the configuration has been seen before
  
  if (Control.output > 2 && myid == 0) {
    cout << "Temporary Connected Cluster:\n";
    for(i=1;i<=length;i++)
      printf("lattice site(%d): (%d, %d, %d, %d)\n",i,
	     configuration[i][1],
	     configuration[i][2],
	     configuration[i][3],
	     configuration[i][4]);	     
  }
  
  findUniqueDescriptorOfCluster(length,
				configuration,
				descriptor);
  
  descriptorIndex=findDescriptorIndex2(descriptor,
				       NumberAnimals);
  
  if (Control.output > 2 && myid == 0)
    cout << "Has it been previously found? =" 
	 << descriptorIndex << endl<<endl;
  
  if ((!descriptorIndex) && (length == maxorder)) {
    updateDescriptors2(descriptor,NumberAnimals);
    NumberAnimals++;
    if (Control.output > 1 && myid == 0) {
      printf("Lattice Animal Number (%d): \n",NumberAnimals);
      printOutDescriptor(descriptor);
      printOutClusterFromDescriptor(descriptor);
      printf("\n");
    }
  }
  
  if (length == maxorder) {

    // this is a `dummy case' whihc just makes sure that we don't
    // go over the maximum LSUBm approximation number adjacent
    // lattice sites for the lattice animals.

  } else {

    for(i=1;i<=length;i++) {
      index=configuration[i][4];
      for(j=1;j<=LatticeNeighbours.number[index];j++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
	tempLength=length;
	
	tempLength++;
	
	configuration[tempLength][1]=configuration[i][1]+
	  LatticeNeighbours.x[index][j];
	configuration[tempLength][2]=configuration[i][2]+
	  LatticeNeighbours.y[index][j];
	configuration[tempLength][3]=configuration[i][3]+
	  LatticeNeighbours.z[index][j];
	configuration[tempLength][4]=LatticeNeighbours.cell[index][j];
	
	tempx=configuration[tempLength][1];
	tempy=configuration[tempLength][2];
	tempz=configuration[tempLength][3];
	tempCell=configuration[tempLength][4];
	
	numberFlips=1;
	
	// check that we have NOT repeated a given site.
	
	for(k=1;k<tempLength;k++)
	  if (tempx==configuration[k][1] &&
	      tempy==configuration[k][2] &&
	      tempz==configuration[k][3] &&
	      tempCell==configuration[k][4])
	    numberFlips++;
	
	if ((numberFlips < 2) && (tempx > -1)) {
	  recursiveClusteringFunction(configuration,tempLength,maxorder,
				      NumberAnimals);
	}
      }
    }
  }
}

void findFundamentalSpinConfigurations(void)
{

  int i,j,k,l;
  int max[MAX_LSUBM+5];
  int max_num_spin_flips;
  int descriptor[MAX_SPIN_LIST+2];
  int temp_descriptor[MAX_SPIN_LIST+2];
  int index1,index2,index3;
  int temp1, temp2, temp3;
  int descriptorIndex;
  int length;
  int **configuration;
  int *spin;
  int **latticeAnimal;
  int tempLength;
  int flag, maxNumberFlips;

  configuration=imatrix(0,MAX_LSUBM+1,0,5);
  spin=ivector(0,BLOCK_LENGTH*MAX_SPIN_LIST+2);
  latticeAnimal=imatrix(0,MAX_LSUBM+1,0,5);

  // for 1D spin-one this is simple.
  
  max_num_spin_flips=((int) (2*Params.s + 1.000001));
  
  max[0]=1;
  for(i=1;i<=BLOCK_LENGTH;i++) {
    max[i]=max[i-1]*max_num_spin_flips;
    if (Control.output > 10 && myid == 0) 
      cout << "max["<<i<<"] = " << max[i] << endl;
    if (max[i] >= 2147438547/max_num_spin_flips) {
      cout << "max[i] >= biggest INTEGER possible...\nExiting!!\n\n";
      exit(1);
    }
  }  

  NumberConfigurations=0;
  
  if (Params.m > (3*BLOCK_LENGTH)) {
    printf("Define: Params.m > (3*BLOCK_LENGTH)...\nExiting!!\n\n");
    exit(1);
  } else if (Params.m > (2*BLOCK_LENGTH)) {
    temp1=Params.m-2*BLOCK_LENGTH;
    temp2=BLOCK_LENGTH;
    temp3=BLOCK_LENGTH;
  } else if (Params.m > BLOCK_LENGTH) {
    temp1=0;
    temp2=Params.m-BLOCK_LENGTH;
    temp3=BLOCK_LENGTH;
  } else {
    temp1=0;
    temp2=0;
    temp3=Params.m;
  }
  
  for(i=1;i<=NumberLatticeAnimals;i++) {
    
    length=findClusterFromDescriptor(latticeAnimals[i].descriptor,
				     latticeAnimal);

    if (length != Params.m) {
      printf("Error in Define: length != Params.m...\nExiting!!\n\n");
      exit(1);
    }

    for(index1=0;index1<max[temp1];index1++) {
      for(index2=0;index2<max[temp2];index2++) {
	for(index3=1;index3<max[temp3];index3++) {
	  
	  for(k=0;k<=MAX_SPIN_LIST;k++) 
	    temp_descriptor[k]=0;
	  
	  temp_descriptor[1]=index3;
	  temp_descriptor[2]=index2;
	  temp_descriptor[3]=index1;
	  
	  findSpinState(temp_descriptor,spin);

	  tempLength=0;
	  flag=0;

	  for(j=1;j<=Params.m;j++) {
	    maxNumberFlips=(int) (2*UnitCell.s[latticeAnimal[j][4]]+0.01);
	    if (spin[j] > maxNumberFlips) 
	      flag=1;
	    for(l=1;l<=spin[j];l++) {
	      tempLength++;
	      configuration[tempLength][1]=latticeAnimal[j][1];
	      configuration[tempLength][2]=latticeAnimal[j][2];
	      configuration[tempLength][3]=latticeAnimal[j][3];
	      configuration[tempLength][4]=latticeAnimal[j][4];
	    }
	  }
	  	  
	  if (tempLength == 0){
	    printf("DEFINE: tempLength == 0...\nExiting!!\n\n");
	    exit(1);
	  }

	  if (flag==0) {

	    findUniqueDescriptorOfCluster(tempLength,
					  configuration,
					  descriptor);
	    
	    descriptorIndex=findDescriptorIndex(descriptor,
						NumberConfigurations);
	    
	    if (Control.output > 2 && myid == 0) {
	      printOutDescriptor(temp_descriptor);
	      printOutClusterFromDescriptor(temp_descriptor);
	      printOutDescriptor(descriptor);
	      printOutClusterFromDescriptor(descriptor);
	      cout << "Has it been previously found? =" 
		   << descriptorIndex << endl;
	      cout<< "Subset == " << subsetClusters(descriptor)<<endl<<endl;
	    }
	    
	    if ((!descriptorIndex) && subsetClusters(descriptor)) {
	      
	      updateDescriptors(descriptor,NumberConfigurations);
	      NumberConfigurations++;
	      
	      if (Control.output > 1 && myid == 0) {
		printOutDescriptor(descriptor);
		printOutClusterFromDescriptor(descriptor);
		printf("\n");
	      }
	    }
	  }
	}
      }
    }
  }

  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  free_imatrix(latticeAnimal,0,MAX_LSUBM+1,0,5);
  free_ivector(spin,0,BLOCK_LENGTH*MAX_SPIN_LIST+2);
}

void obtainDisconnectedClusters(int &NumberDisconnected, 
				int NumberConnected)
{
}

void updateDescriptors(int *descriptor,
		       int numberElements)
{

  int i,j,k,l;
  int position;
  int tempLogicFlag;
  
  // first find where to put the data element 
  // just use a simple backwards linear search

  // although this bit is as cumbersome as old routine it is called
  // less often and all of the elements are always sorted!

  position=1;

  if (numberElements > 0) {
    for(j=numberElements;j>0;j--) {
      tempLogicFlag=0;
      for(i=MAX_SPIN_LIST-1;i>0;i--) {
	if (spinConfigurations[j].descriptor[i]<descriptor[i]) {
	  tempLogicFlag=1;
	  break;
	} else if (spinConfigurations[j].descriptor[i]>descriptor[i]){
	  tempLogicFlag=2;
	  break;
	}
      }      
      if (tempLogicFlag==1) {
	position=j+1;
	break;
      }
    }
  }
  
  // now put the element in the correct position
  
  if (numberElements > 0) {
    for(j=numberElements;j>=position;j--) {
      for(i=1;i<MAX_SPIN_LIST;i++) {
	spinConfigurations[j+1].descriptor[i]=
	  spinConfigurations[j].descriptor[i];
      }
    }
  }
  
  for(i=1;i<MAX_SPIN_LIST;i++)
    spinConfigurations[position].descriptor[i]=descriptor[i];
  
  if (Control.output > 3 && myid == 0) {
    for(j=1;j<=numberElements+1;j++) {
      for(i=1;i<MAX_SPIN_LIST;i++)
	cout << spinConfigurations[j].descriptor[i] << " ";
      cout << endl;
    }
    cout <<endl;
  }
}

void updateDescriptors2(int *descriptor,
			int numberElements)
{
  
  int i,j,k,l;
  int position;
  int tempLogicFlag;
  
  // first find where to put the data element 
  // just use a simple backwards linear search

  // although this bit is as cumbersome as old routine it is called
  // less often and all of the elements are always sorted!

  position=1;

  if (numberElements > 0) {
    for(j=numberElements;j>0;j--) {
      tempLogicFlag=0;
      for(i=MAX_SPIN_LIST-1;i>0;i--) {
	if (latticeAnimals[j].descriptor[i]<descriptor[i]) {
	  tempLogicFlag=1;
	  break;
	} else if (latticeAnimals[j].descriptor[i]>descriptor[i]){
	  tempLogicFlag=2;
	  break;
	}
      }      
      if (tempLogicFlag==1) {
	position=j+1;
	break;
      }
    }
  }
  
  // now put the element in the correct position
  
  if (numberElements > 0) {
    for(j=numberElements;j>=position;j--) {
      for(i=1;i<MAX_SPIN_LIST;i++) {
	latticeAnimals[j+1].descriptor[i]=
	  latticeAnimals[j].descriptor[i];
      }
    }
  }
  
  for(i=1;i<MAX_SPIN_LIST;i++)
    latticeAnimals[position].descriptor[i]=descriptor[i];
  
  if (Control.output > 3 && myid == 0) {
    for(j=1;j<=numberElements+1;j++) {
      for(i=1;i<MAX_SPIN_LIST;i++)
	cout << latticeAnimals[j].descriptor[i] << " ";
      cout << endl;
    }
    cout <<endl;
  }
}

int subsetClusters(int *descriptor)
{

  int i,j,k,l;
  int length;
  int **config;
  int flag;
  int temp;
  double x, y, z,stz;
  int int_x, int_y, int_z;

  config=imatrix(0,MAX_LSUBM+1,0,5);

  // this routine is somewhat lattice specific and is used to 
  // restrict the clusters used to within some well-defined
  // subset.

  // Hence, this subroutine is application-dependent and to be 
  // written by the user.

  // it returns a value of 0 if the cluster is acceptable
  // and a non-zero integer otherwise.

  length=findClusterFromDescriptor(descriptor,config);

  flag=1;

  // always check that we don't have more than n spin flips.

  if (length > Params.n)
    flag=0; 

  // restriction 1: NO RESTRICTION!

  // restriction 2: even number of spin flips.

  if (restrictionClusters==2) {
    if (flag==1) {
      if (length%2!=0)
	flag=0;
    }
  }
  
  // restriction 3: S_T^z=0

  if (restrictionClusters==3) {

    // USER DEFINED OPTION!!

    // ADVANCED: USER MUST ALTER CODE IN THIS CASE!!!

    // EXAMPLE: XXZ model or Heisenberg linear-chain, square or cubic
    // lattices arbitrary `s' though.

    // EXAMPLE: Ca V_4 O_9 Neel model state. Take out comment markers
    // on `x+=0.5;' and `y+=0.5;' and recompile

    if (flag==1) {
      stz=0.0;
      for(i=1;i<=length;i++) {
	findCartesiansFromGeneralisedPosition(x,y,z,config[i]);
	// for CAVO applications and SRCUBO now too!

	   x+=0.5000000;
	   y+=0.5000000;

	if (((int)(((fabs(x)+fabs(y)+fabs(z))+0.001)/2.0)) == 
	    ((int)(((fabs(x)+fabs(y)+fabs(z))+1.001)/2.0)))
	// For CAVO with striped model state 
	//	if (((int)(((fabs(y)+0.001)/2.0))) == ((int)((fabs(y)+1.001)/2.0)))
	  stz++;
	else
	  stz--;
      }
      if (fabs(stz)>0.1)
	flag=0;
    }
  }

  free_imatrix(config,0,MAX_LSUBM+1,0,5);
  return flag;

}

void findSpinState(int *descriptor, int *spinState) 
{

  int i,j,k,l;
  int max[MAX_LSUBM+5];
  int max_num_spin_flips;
  int temp_descriptor[MAX_SPIN_LIST+2];
  int index1,index2,index3;
  int temp,y;

  max_num_spin_flips=((int) (2*Params.s + 1.000001));
  
  max[0]=1;
  for(i=1;i<=BLOCK_LENGTH;i++) {
    max[i]=max[i-1]*max_num_spin_flips;
    if (Control.output > 10 && myid == 0) 
      cout << "max["<<i<<"] = " << max[i] << endl;
    if (max[i] >= 2147438547/max_num_spin_flips) {
      cout << "max[i] >= biggest INTEGER possible...\nExiting!!\n\n";
      exit(1);
    }
  }

  for(i=1;i<=MAX_SPIN_LIST;i++) {
    temp=descriptor[i];
    for(j=BLOCK_LENGTH;j>0;j--) {
      y=temp/max[j-1];
      spinState[(i-1)*BLOCK_LENGTH+j]=y;
      temp=temp-y*max[j-1];
      //      cout << "temp = " << temp << endl;
    }
  }
}

