

struct prm {

  /* This structure contains all of the globally visible information
     concerning runtime CCM Parameters set by the users either in the
     code itself, as command line instructions, using the GUI, or in
     some input file. 
     
     This is distinct from Control Parameters which simply tell the 
     code how much output to make, what filenames to use. 
     
  */


  int n;                // SUBn-m (n) 
  int m;                // SUBn-m (m) 
  int btLength;         // Determines size of Bounding Traingle
  int dim;              // Dimension Of Lattice
  int nFundG;           // # Fund. Configs G-State
  int nFundE;           // # Fund. Configs E-State
  float s;              // Global Quantum Spin Number
  // NOTE Params.s MUST be set as big as the largest .s in the UNIT CELL!!!
  int ddd;

};

int  max_number_cells_BT;

struct ctrl {
  /* 
     
     Control Parameters which simply tell the 
     code how much output to make, what filenames to use. 
     
  */

  int output;           // Level Of Output
  int file;             // save CCM equations to file
  char name[60];        // Root File Name                

};

prm Params;

ctrl Control;

struct UC {

  /* this structure completely defines the unit cell */

  int number;             // number of sites in unit cell (UC)
  double x[MAX_UNIT_CELL];   // 
  double y[MAX_UNIT_CELL];   // number of sites in 
  double z[MAX_UNIT_CELL];   // number of sites in 
  float s[MAX_UNIT_CELL]; // `local' quantum spin number for each site in UC
  
};

UC UnitCell;

struct neighbours {

  /* this is an interesting structure because it defines the 
     nearest neighbour structure for each of the elements of 
     the unit cell.

     Essentially each unit cell has however many nearest-neighbours,
     some of which in within the unit cell itself, and some which
     are in neigbouring cells.

     This information is most easily coded in by the user, or
     by writing a simple bit of code to determine the neighbours
     matrix of the lattice or Hamiltonian.

     Note that this also offers the possiblity of defining
     nearest-neighbours, next-nearest-neighbours, next-next-...
     and so on.

  */

  int number[MAX_UNIT_CELL]; /* number of nearest-neighbours for
				each site in the UC */
  int x[MAX_UNIT_CELL][MAX_NUM_NEIGHBOURS]; // TRANSLATION x bravais
  int y[MAX_UNIT_CELL][MAX_NUM_NEIGHBOURS]; // TRANSLATION y bravais
  int z[MAX_UNIT_CELL][MAX_NUM_NEIGHBOURS]; // TRANSLATION z bravais
  int cell[MAX_UNIT_CELL][MAX_NUM_NEIGHBOURS]; // which site in UC
  
};

neighbours LatticeNeighbours_k_to_m;
neighbours LatticeNeighbours_m_to_k;
neighbours LatticeNeighbours;

struct bravais {

  /* structure defines the bavais lattice */

  double a[4];
  double b[4];
  double c[4];

};

bravais BravaisLattice;

struct symmetries {

  /* this structure contains all of the information concerning
     the symmetries of the lattice, Hamiltonian, unit cell etc. */

  int number;     // number of symmetries in lattice, Hamiltonian etc
  int switchList[MAX_NUM_SYM]; // which symmetries from function switch(int)
  int switchFunction(int switchVariable, 
		     double x1, 
		     double y1, 
		     double z1,
		     double &x2, 
		     double &y2,
		     double &z2);
};

symmetries LatticeSymmetries;


struct spinConfigs {
  
  /* 

     This structure describes a particular spin configuration by storing 
     the information in an array. The size of the (integer) numbers
     becomes too long for saving as a standard `int' -- or even `long int.'
     
     Each integer element gives the information on 20 lattices sites
     each of which has a spin of quantum spin number Params.s or
     UnitCell.s[]
     
  */
  
  int descriptor[MAX_SPIN_LIST];
  
};

spinConfigs spinConfigurations[MAX_NUM_CONFIGURATIONS];
spinConfigs connectedConfigurations[MAX_NUM_CONFIGURATIONS];
int         NumberConfigurations;

double CCMcoeffs[MAX_NUM_CONFIGURATIONS];

struct hamiltonian_coefficients{
  double coefficient; 
};

hamiltonian_coefficients Hamiltonian[60];

int lookUpTable[MAX_BT+1][5];
int BTarray[MAX_LSUBM][MAX_LSUBM][MAX_LSUBM];

double x_array[200000];
double y_array[200000];
double z_array[200000];
int    x_index[200000];
int    y_index[200000];
int    z_index[200000];
int    cell_index[200000];
int    totalPoints;
